@@ -1,5 +1,6 @@
 import json
 import os
+import re
 import time
 from collections import namedtuple
 from datetime import datetime
@@ -8,11 +9,11 @@ from metaflow.exception import MetaflowInternalError
 from metaflow.util import get_username, resolve_identity
 
 
-DataArtifact = namedtuple('DataArtifact',
-                          'name ds_type url type sha')
+DataArtifact = namedtuple("DataArtifact", "name ds_type ds_root url type sha")
 
-MetaDatum = namedtuple('MetaDatum',
-                       'field value type tags')
+MetaDatum = namedtuple("MetaDatum", "field value type tags")
+
+attempt_id_re = re.compile(r"attempt_id:([0-9]+)")
 
 
 class MetadataProviderMeta(type):
@@ -39,18 +40,18 @@ def with_metaclass(mcls):
     def decorator(cls):
         body = vars(cls).copy()
         # clean out class body
-        body.pop('__dict__', None)
-        body.pop('__weakref__', None)
+        body.pop("__dict__", None)
+        body.pop("__weakref__", None)
         return mcls(cls.__name__, cls.__bases__, body)
+
     return decorator
 
 
 @with_metaclass(MetadataProviderMeta)
 class MetadataProvider(object):
-
     @classmethod
     def compute_info(cls, val):
-        '''
+        """
         Compute the new information for this provider
 
         The computed value should be returned and will then be accessible directly as cls.INFO.
@@ -66,12 +67,12 @@ class MetadataProvider(object):
         -------
         str :
             Value to be set to INFO
-        '''
-        return ''
+        """
+        return ""
 
     @classmethod
     def default_info(cls):
-        '''
+        """
         Returns the default information for this provider
 
         This should compute and return the default value for the information regarding this provider.
@@ -81,22 +82,22 @@ class MetadataProvider(object):
         -------
         str
             Value to be set by default in INFO
-        '''
-        return ''
+        """
+        return ""
 
     def version(self):
-        '''
+        """
         Returns the version of this provider
 
         Returns
         -------
         str
             Version of the provider
-        '''
-        return ''
+        """
+        return ""
 
-    def new_run_id(self, tags=[], sys_tags=[]):
-        '''
+    def new_run_id(self, tags=None, sys_tags=None):
+        """
         Creates an ID and registers this new run.
 
         The run ID will be unique within a given flow.
@@ -104,19 +105,19 @@ class MetadataProvider(object):
         Parameters
         ----------
         tags : list, optional
-            Tags to apply to this particular run, by default []
+            Tags to apply to this particular run, by default None
         sys_tags : list, optional
-            System tags to apply to this particular run, by default []
+            System tags to apply to this particular run, by default None
 
         Returns
         -------
         int
             Run ID for the run
-        '''
+        """
         raise NotImplementedError()
 
-    def register_run_id(self, run_id, tags=[], sys_tags=[]):
-        '''
+    def register_run_id(self, run_id, tags=None, sys_tags=None):
+        """
         No-op operation in this implementation.
 
         Parameters
@@ -124,14 +125,14 @@ class MetadataProvider(object):
         run_id : int
             Run ID for this run
         tags : list, optional
-            Tags to apply to this particular run, by default []
+            Tags to apply to this particular run, by default None
         sys_tags : list, optional
-            System tags to apply to this particular run, by default []
-        '''
+            System tags to apply to this particular run, by default None
+        """
         raise NotImplementedError()
 
-    def new_task_id(self, run_id, step_name, tags=[], sys_tags=[]):
-        '''
+    def new_task_id(self, run_id, step_name, tags=None, sys_tags=None):
+        """
         Creates an ID and registers this new task.
 
         The task ID will be unique within a flow, run and step
@@ -143,19 +144,21 @@ class MetadataProvider(object):
         step_name : string
             Name of the step
         tags : list, optional
-            Tags to apply to this particular task, by default []
+            Tags to apply to this particular task, by default None
         sys_tags : list, optional
-            System tags to apply to this particular task, by default []
+            System tags to apply to this particular task, by default None
 
         Returns
         -------
         int
             Task ID for the task
-        '''
+        """
         raise NotImplementedError()
 
-    def register_task_id(self, run_id, step_name, task_id, tags=[], sys_tags=[]):
-        '''
+    def register_task_id(
+        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
+    ):
+        """
         No-op operation in this implementation.
 
         Parameters
@@ -170,11 +173,11 @@ class MetadataProvider(object):
             Tags to apply to this particular run, by default []
         sys_tags : list, optional
             System tags to apply to this particular run, by default []
-        '''
+        """
         raise NotImplementedError()
 
     def get_runtime_environment(self, runtime_name):
-        '''
+        """
         Returns a dictionary of environment variables to be set
 
         Parameters
@@ -186,17 +189,13 @@ class MetadataProvider(object):
         -------
         dict[string] -> string
             Environment variables from this metadata provider
-        '''
-        return {'METAFLOW_RUNTIME_NAME': runtime_name,
-                'USER': get_username()}
+        """
+        return {"METAFLOW_RUNTIME_NAME": runtime_name, "USER": get_username()}
 
-    def register_data_artifacts(self,
-                                run_id,
-                                step_name,
-                                task_id,
-                                attempt_id,
-                                artifacts):
-        '''
+    def register_data_artifacts(
+        self, run_id, step_name, task_id, attempt_id, artifacts
+    ):
+        """
         Registers the fact that the data-artifacts are associated with
         the particular task.
 
@@ -215,11 +214,11 @@ class MetadataProvider(object):
             Attempt for the task
         artifacts : List of DataArtifact
             Artifacts associated with this task
-        '''
+        """
         raise NotImplementedError()
 
     def register_metadata(self, run_id, step_name, task_id, metadata):
-        '''
+        """
         Registers metadata with a task.
 
         Note that the same metadata can be registered multiple times for the same task (for example
@@ -236,7 +235,7 @@ class MetadataProvider(object):
             Task ID for the task
         metadata : List of MetaDatum
             Metadata associated with this task
-        '''
+        """
         raise NotImplementedError()
 
     def start_task_heartbeat(self, flow_id, run_id, step_name, task_id):
@@ -249,8 +248,10 @@ class MetadataProvider(object):
         pass
 
     @classmethod
-    def _get_object_internal(cls, obj_type, obj_order, sub_type, sub_order, filters=None, *args):
-        '''
+    def _get_object_internal(
+        cls, obj_type, obj_order, sub_type, sub_order, filters, attempt, *args
+    ):
+        """
         Return objects for the implementation of this class
 
         See get_object_internal for the description of what this function does
@@ -269,16 +270,23 @@ class MetadataProvider(object):
             Dictionary with keys 'any_tags', 'tags' and 'system_tags'. If specified
             will return only objects that have the specified tags present. Filters
             are ANDed together so all tags must be present for the object to be returned.
+        attempt : int or None
+            If None, returns artifacts for latest *done* attempt and all metadata. Otherwise,
+            returns artifacts for that attempt (existent, done or not) and *all* metadata
+            NOTE: Unlike its external facing `get_object`, this function should
+            return *all* metadata; the base class will properly implement the
+            filter. For artifacts, this function should filter artifacts at
+            the backend level.
 
         Return
         ------
             object or list :
                 Depending on the call, the type of object return varies
-        '''
+        """
         raise NotImplementedError()
 
-    def add_sticky_tags(self, tags=[], sys_tags=[]):
-        '''
+    def add_sticky_tags(self, tags=None, sys_tags=None):
+        """
         Adds tags to be added to every run and task
 
         Tags can be added to record information about a run/task. Such tags can be specified on a
@@ -289,16 +297,18 @@ class MetadataProvider(object):
         Parameters
         ----------
         tags : list, optional
-            Tags to add to every run/task, by default []
+            Tags to add to every run/task, by default None
         sys_tags : list, optional
-            System tags to add to every run/task, by default []
-        '''
-        self.sticky_tags.extend(tags)
-        self.sticky_sys_tags.extend(sys_tags)
+            System tags to add to every run/task, by default None
+        """
+        if tags:
+            self.sticky_tags.update(tags)
+        if sys_tags:
+            self.sticky_sys_tags.update(sys_tags)
 
     @classmethod
-    def get_object(cls, obj_type, sub_type, filters=None, *args):
-        '''Returns the requested object depending on obj_type and sub_type
+    def get_object(cls, obj_type, sub_type, filters, attempt, *args):
+        """Returns the requested object depending on obj_type and sub_type
 
         obj_type can be one of 'root', 'flow', 'run', 'step', 'task',
         or 'artifact'
@@ -330,89 +340,123 @@ class MetadataProvider(object):
             Dictionary with keys 'any_tags', 'tags' and 'system_tags'. If specified
             will return only objects that have the specified tags present. Filters
             are ANDed together so all tags must be present for the object to be returned.
+        attempt : int or None
+            If None, for metadata and artifacts:
+              - returns information about the latest attempt for artifacts
+              - returns all metadata across all attempts
+            Otherwise, returns information about metadata and artifacts for that
+            attempt only.
+            NOTE: For older versions of Metaflow (pre 2.4.0), the attempt for
+            metadata is not known; in that case, all metadata is returned (as
+            if None was passed in).
 
         Return
         ------
             object or list :
                 Depending on the call, the type of object return varies
-        '''
+        """
         obj_order = {
-            'root': 0,
-            'flow': 1,
-            'run': 2,
-            'step': 3,
-            'task': 4,
-            'artifact': 5,
-            'metadata': 6,
-            'self': 7}
+            "root": 0,
+            "flow": 1,
+            "run": 2,
+            "step": 3,
+            "task": 4,
+            "artifact": 5,
+            "metadata": 6,
+            "self": 7,
+        }
         type_order = obj_order.get(obj_type)
         sub_order = obj_order.get(sub_type)
 
         if type_order is None:
-            raise MetaflowInternalError(msg='Cannot find type %s' % obj_type)
+            raise MetaflowInternalError(msg="Cannot find type %s" % obj_type)
         if type_order > 5:
-            raise MetaflowInternalError(msg='Type %s is not allowed' % obj_type)
+            raise MetaflowInternalError(msg="Type %s is not allowed" % obj_type)
 
         if sub_order is None:
-            raise MetaflowInternalError(msg='Cannot find subtype %s' % sub_type)
+            raise MetaflowInternalError(msg="Cannot find subtype %s" % sub_type)
 
         if type_order >= sub_order:
-            raise MetaflowInternalError(msg='Subtype %s not allowed for %s' % (sub_type, obj_type))
+            raise MetaflowInternalError(
+                msg="Subtype %s not allowed for %s" % (sub_type, obj_type)
+            )
 
         # Metadata is always only at the task level
-        if sub_type == 'metadata' and obj_type != 'task':
-            raise MetaflowInternalError(msg='Metadata can only be retrieved at the task level')
+        if sub_type == "metadata" and obj_type != "task":
+            raise MetaflowInternalError(
+                msg="Metadata can only be retrieved at the task level"
+            )
 
-        return cls._get_object_internal(obj_type, type_order, sub_type, sub_order, filters, *args)
+        if attempt is not None:
+            try:
+                attempt_int = int(attempt)
+                if attempt_int < 0:
+                    raise ValueError("Attempt can only be positive")
+            except ValueError:
+                raise ValueError("Attempt can only be a positive integer")
+        else:
+            attempt_int = None
 
-    def _all_obj_elements(self, tags=[], sys_tags=[]):
+        pre_filter = cls._get_object_internal(
+            obj_type, type_order, sub_type, sub_order, filters, attempt_int, *args
+        )
+        if attempt_int is None or sub_order != 6:
+            # If no attempt or not for metadata, just return as is
+            return pre_filter
+        return MetadataProvider._reconstruct_metadata_for_attempt(
+            pre_filter, attempt_int
+        )
+
+    def _all_obj_elements(self, tags=None, sys_tags=None):
         user = get_username()
         return {
-            'flow_id': self._flow_name,
-            'user_name': user,
-            'tags': tags,
-            'system_tags': sys_tags,
-            'ts_epoch': int(round(time.time() * 1000))}
+            "flow_id": self._flow_name,
+            "user_name": user,
+            "tags": list(tags) if tags else [],
+            "system_tags": list(sys_tags) if sys_tags else [],
+            "ts_epoch": int(round(time.time() * 1000)),
+        }
 
     def _flow_to_json(self):
         # No need to store tags, sys_tags or username at the flow level
         # since runs are the top level logical concept, which is where we
         # store tags, sys_tags and username
-        return {
-            'flow_id': self._flow_name,
-            'ts_epoch': int(round(time.time() * 1000))}
+        return {"flow_id": self._flow_name, "ts_epoch": int(round(time.time() * 1000))}
 
-    def _run_to_json(self, run_id=None, tags=[], sys_tags=[]):
+    def _run_to_json(self, run_id=None, tags=None, sys_tags=None):
         if run_id is not None:
-            d = {'run_number': run_id}
+            d = {"run_number": run_id}
         else:
             d = {}
         d.update(self._all_obj_elements(tags, sys_tags))
         return d
 
-    def _step_to_json(self, run_id, step_name, tags=[], sys_tags=[]):
-        d = {
-            'run_number': run_id,
-            'step_name': step_name}
+    def _step_to_json(self, run_id, step_name, tags=None, sys_tags=None):
+        d = {"run_number": run_id, "step_name": step_name}
         d.update(self._all_obj_elements(tags, sys_tags))
         return d
 
-    def _task_to_json(self, run_id, step_name, task_id=None, tags=[], sys_tags=[]):
-        d = {
-            'run_number': run_id,
-            'step_name': step_name}
+    def _task_to_json(self, run_id, step_name, task_id=None, tags=None, sys_tags=None):
+        d = {"run_number": run_id, "step_name": step_name}
         if task_id is not None:
-            d['task_id'] = task_id
+            d["task_id"] = task_id
         d.update(self._all_obj_elements(tags, sys_tags))
         return d
 
     def _object_to_json(
-            self, obj_type, run_id=None, step_name=None, task_id=None, tags=[], sys_tags=[]):
-        if obj_type == 'task':
+        self,
+        obj_type,
+        run_id=None,
+        step_name=None,
+        task_id=None,
+        tags=None,
+        sys_tags=None,
+    ):
+        if obj_type == "task":
             return self._task_to_json(run_id, step_name, task_id, tags, sys_tags)
-        if obj_type == 'step':
+        if obj_type == "step":
             return self._step_to_json(run_id, step_name, tags, sys_tags)
-        if obj_type == 'run':
+        if obj_type == "run":
             return self._run_to_json(run_id, tags, sys_tags)
         return self._flow_to_json()
 
@@ -420,60 +464,73 @@ class MetadataProvider(object):
         result = []
         for art in artifacts:
             d = {
-                'run_number': run_id,
-                'step_name': step_name,
-                'task_id': task_id,
-                'attempt_id': attempt_id,
-                'name': art.name,
-                'content_type': art.type,
-                'type': 'metaflow.artifact',
-                'sha': art.sha,
-                'ds_type': art.ds_type,
-                'location': art.url}
+                "run_number": run_id,
+                "step_name": step_name,
+                "task_id": task_id,
+                "attempt_id": attempt_id,
+                "name": art.name,
+                "content_type": art.type,
+                "type": "metaflow.artifact",
+                "sha": art.sha,
+                "ds_type": art.ds_type,
+                "location": art.url if art.url else ":root:%s" % art.ds_root,
+            }
             d.update(self._all_obj_elements(self.sticky_tags, self.sticky_sys_tags))
             result.append(d)
         return result
 
     def _metadata_to_json(self, run_id, step_name, task_id, metadata):
         user = get_username()
-        return [{
-            'flow_id': self._flow_name,
-            'run_number': run_id,
-            'step_name': step_name,
-            'task_id': task_id,
-            'field_name': datum.field,
-            'type': datum.type,
-            'value': datum.value,
-            'tags': datum.tags,
-            'user_name': user,
-            'ts_epoch': int(round(time.time() * 1000))} for datum in metadata]
+        return [
+            {
+                "flow_id": self._flow_name,
+                "run_number": run_id,
+                "step_name": step_name,
+                "task_id": task_id,
+                "field_name": datum.field,
+                "type": datum.type,
+                "value": datum.value,
+                "tags": list(set(datum.tags)) if datum.tags else [],
+                "user_name": user,
+                "ts_epoch": int(round(time.time() * 1000)),
+            }
+            for datum in metadata
+        ]
 
     def _tags(self):
         env = self._environment.get_environment_info()
         tags = [
             resolve_identity(),
-            'runtime:' + env['runtime'],
-            'python_version:' + env['python_version_code'],
-            'date:' + datetime.utcnow().strftime('%Y-%m-%d')]
-        if env['metaflow_version']:
-            tags.append('metaflow_version:' + env['metaflow_version'])
-        if 'metaflow_r_version' in env:
-            tags.append('metaflow_r_version:' + env['metaflow_r_version'])
-        if 'r_version_code' in env:
-            tags.append('r_version:' + env['r_version_code'])
+            "runtime:" + env["runtime"],
+            "python_version:" + env["python_version_code"],
+            "date:" + datetime.utcnow().strftime("%Y-%m-%d"),
+        ]
+        if env["metaflow_version"]:
+            tags.append("metaflow_version:" + env["metaflow_version"])
+        if "metaflow_r_version" in env:
+            tags.append("metaflow_r_version:" + env["metaflow_r_version"])
+        if "r_version_code" in env:
+            tags.append("r_version:" + env["r_version_code"])
         return tags
 
-    def _register_code_package_metadata(self, run_id, step_name, task_id):
+    def _register_code_package_metadata(self, run_id, step_name, task_id, attempt):
         metadata = []
-        code_sha = os.environ.get('METAFLOW_CODE_SHA')
-        code_url = os.environ.get('METAFLOW_CODE_URL')
-        code_ds = os.environ.get('METAFLOW_CODE_DS')
+        code_sha = os.environ.get("METAFLOW_CODE_SHA")
+        code_url = os.environ.get("METAFLOW_CODE_URL")
+        code_ds = os.environ.get("METAFLOW_CODE_DS")
         if code_sha:
-            metadata.append(MetaDatum(
-                field='code-package',
-                value=json.dumps({'ds_type': code_ds, 'sha': code_sha, 'location': code_url}),
-                type='code-package',
-                tags=[]))
+            metadata.append(
+                MetaDatum(
+                    field="code-package",
+                    value=json.dumps(
+                        {"ds_type": code_ds, "sha": code_sha, "location": code_url}
+                    ),
+                    type="code-package",
+                    tags=["attempt_id:{0}".format(attempt)],
+                )
+            )
+        # We don't tag with attempt_id here because not readily available; this
+        # is ok though as this doesn't change from attempt to attempt.
         if metadata:
             self.register_metadata(run_id, step_name, task_id, metadata)
 
@@ -484,30 +541,67 @@ class MetadataProvider(object):
         starting_point = elts
         result = []
         for key, value in filters.items():
-            if key == 'any_tags':
+            if key == "any_tags":
                 for obj in starting_point:
-                    if value in obj.get('tags', []) or value in obj.get('system_tags', []):
+                    if value in obj.get("tags", []) or value in obj.get(
+                        "system_tags", []
+                    ):
                         result.append(obj)
-            if key == 'tags':
+            if key == "tags":
                 for obj in starting_point:
-                    if value in obj.get('tags', []):
+                    if value in obj.get("tags", []):
                         result.append(obj)
-            if key == 'system_tags':
+            if key == "system_tags":
                 for obj in starting_point:
-                    if value in obj.get('system_tags', []):
+                    if value in obj.get("system_tags", []):
                         result.append(obj)
             starting_point = result
             result = []
         return starting_point
 
+    @staticmethod
+    def _reconstruct_metadata_for_attempt(all_metadata, attempt_id):
+        have_all_attempt_id = True
+        attempts_start = {}
+        post_filter = []
+        for v in all_metadata:
+            if v["field_name"] == "attempt":
+                attempts_start[int(v["value"])] = v["ts_epoch"]
+            all_tags = v.get("tags")
+            if all_tags is None:
+                all_tags = []
+            for t in all_tags:
+                match_result = attempt_id_re.match(t)
+                if match_result:
+                    if int(match_result.group(1)) == attempt_id:
+                        post_filter.append(v)
+                    break
+            else:
+                # We didn't encounter a match for attempt_id
+                have_all_attempt_id = False
+
+        if not have_all_attempt_id:
+            # We reconstruct base on the attempts_start
+            start_ts = attempts_start.get(attempt_id, -1)
+            if start_ts < 0:
+                return []  # No metadata since the attempt hasn't started
+            # Doubt we will be using Python in year 3000
+            end_ts = attempts_start.get(attempt_id + 1, 32503680000000)
+            post_filter = [
+                v
+                for v in all_metadata
+                if v["ts_epoch"] >= start_ts and v["ts_epoch"] < end_ts
+            ]
+
+        return post_filter
+
     def __init__(self, environment, flow, event_logger, monitor):
         self._task_id_seq = -1
-        self.sticky_tags = []
-        self.sticky_sys_tags = []
+        self.sticky_tags = set()
+        self.sticky_sys_tags = set()
         self._flow_name = flow.name
         self._event_logger = event_logger
         self._monitor = monitor
         self._environment = environment
-        self._runtime = os.environ.get(
-            'METAFLOW_RUNTIME_NAME', 'dev')
+        self._runtime = os.environ.get("METAFLOW_RUNTIME_NAME", "dev")
         self.add_sticky_tags(sys_tags=self._tags())
