@@ -7,15 +7,17 @@ except:
 
 from .exception import MetaflowException
 
+
 class PyLintWarn(MetaflowException):
-    headline="Pylint is not happy"
+    headline = "Pylint is not happy"
+
 
 class PyLint(object):
-
     def __init__(self, fname):
         self._fname = fname
         try:
             from pylint.lint import Run
+
             self._run = Run
         except:
             self._run = None
@@ -26,7 +28,7 @@ class PyLint(object):
     def run(self, logger=None, warnings=False, pylint_config=[]):
         args = [self._fname]
         if not warnings:
-            args.append('--errors-only')
+            args.append("--errors-only")
         if pylint_config:
             args.extend(pylint_config)
         stdout = sys.stdout
@@ -50,14 +52,14 @@ class PyLint(object):
             warnings = True
 
         if warnings:
-            raise PyLintWarn('*Fix Pylint warnings listed above or say --no-pylint.*')
+            raise PyLintWarn("*Fix Pylint warnings listed above or say --no-pylint.*")
 
         return pylint_is_happy, pylint_exception_msg
 
     def _filter_lines(self, output):
         for line in output.splitlines():
             # Ignore headers
-            if '***' in line:
+            if "***" in line:
                 continue
             # Ignore complaints about decorators missing in the metaflow module.
             # Automatic generation of decorators confuses Pylint.
